/*  Vexed v2.0 Beta - gfx.c "The graphical functions for Vexed"
    Copyright (C) 1999 James McCombe (cybertube@earthling.net)
    
    August 17, 2001 - Version 2.0 changes by Mark Ingebretson
    (ingebret@yahoo.com) and others.
    
    This file is part of Vexed.

    Vexed is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Vexed is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Vexed; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "PalmOS.h"

#include "vexed.h"
#include "protos.h"

extern PlayGround    current;
extern PlayGroundIntro  currentI;
extern Preferences   VexedPreferences;
extern Boolean       colorOk;
extern BitmapPtr     ObjectBitmapPtr[BitmapCount];
extern WinHandle     ObjectWindowHandles[BitmapCount];
extern Boolean       InBlockCheck;           // are we in a block check
extern Boolean       BlockCheckPending;      // waiting on a block check
extern UInt8         BlockCheckX,            // x,y of pending block check
                     BlockCheckY;
extern Boolean       InSolution;             // Are we displaying a solution
extern Boolean       InSolutionBack;         // are we doing a back move in solution?
extern Boolean       MoveNum;                // Which move of solution are we now displaying
extern Boolean       SolutionRun;            // Is solution in free-running mode

extern char gszEllipsis[4];

//-----------------------------------------------------------------------------
void DrawBlinds() {
   Int16             i, x;
   RectangleType     r;

   if (VexedPreferences.BlindsEffect && !InSolutionBack) {
      r.extent.y = 128;
      r.topLeft.y = BlockSize;

      for (i = 0; i <= BlockSize; i ++) {
         r.extent.x = i;
         for (x = 0; x <= 144; x += BlockSize) {
            r.topLeft.x = x;
            WinEraseRectangle(&r, 0);
         }
         SysTaskDelay(1);
      }
   }
   return;
}

//-----------------------------------------------------------------------------
void RedrawLevel() {
   Int8              x, y;
   RectangleType     r;

   // erase the whole screen first (Air)
   r.topLeft.x = 0;
   r.topLeft.y = 16;
   r.extent.x = sizeX * BlockSize;
   r.extent.y = sizeY * BlockSize;
   WinEraseRectangle(&r, 0);

   // set up bitmap rectangle copies
   // r.topLeft.x = 0;                       // already 0
   r.topLeft.y = 0;
   r.extent.x = r.extent.y = BlockSize;
   for (y = 0; y < sizeY; y++) {
      for (x = 0; x < sizeX; x++) {
         if (current[y][x] != Air) {
            //WinDrawBitmap(Bitmap[current[y][x]], x * BlockSize, y * BlockSize + 16);
            //WinCopyRectangle (ObjectWindowHandles[current[y][x] - 1], 0, &r, 
            //                  x * BlockSize, y * BlockSize + 16, winPaint);
            DrawBM(current[y][x] - 1, x * BlockSize, y * BlockSize + 16, false);

         }
      }
   }
   UpdateStats();                            // display the on screen stats
   return;
}

//-----------------------------------------------------------------------------
void DrawLevel() {
   // do the cool blinds effect
   DrawBlinds(); 
   // Redraw Level
   RedrawLevel();
   return;
}

#define kcxStat 160
#define kcxBorder 3

//-----------------------------------------------------------------------------
void UpdateStats() {
	char szTitle[64];
	char szMoves[32];
	char szScore[32];
	char szScoreInteger[8];
	char sz[128];
	int cx, cy;
	UInt32 dwVer;
	UInt32 rgbBack, rgbFore;
	UInt32 rgbBackSav, rgbForeSav;
	FontID fnt, fntSav;
	RectangleType rc;
	int nScore;
	int cxScore, cyScore;
	int cxScoreInteger;

   if (InSolution)
		return;

	// Get the title and move count strings

   if (GetLevelProperty(VexedPreferences.LevelPack[0].CurrentLevel, "title", sz, sizeof(sz))) {
		StrPrintF(szTitle, "%d: %s", VexedPreferences.LevelPack[0].CurrentLevel, sz);
   } else {
		StrPrintF(szTitle, "Level %d:", VexedPreferences.LevelPack[0].CurrentLevel);
   }
	if (VexedPreferences.LevelPack[0].MoveCountPar == -1) {
		szMoves[0] = 0;
	} else {
		StrPrintF(szMoves, " %d/%d", VexedPreferences.LevelPack[0].MoveCountMade, VexedPreferences.LevelPack[0].MoveCountPar);
	}
	StrCopy(sz, szTitle);
	StrCat(sz, szMoves);

	// Get the score text

	nScore = CalcScore(0);
	StrIToA(szScoreInteger, nScore);
	StrCopy(szScore, "Score ");
	StrCat(szScore, szScoreInteger);

	// First figure out which font we'll use. Try to use bold font - looks better

	fnt = boldFont;
	fntSav = FntSetFont(fnt);
	cx = FntCharsWidth(sz, StrLen(sz));
	cxScore = FntCharsWidth(szScore, StrLen(szScore));
	cxScoreInteger = FntCharsWidth(szScoreInteger, StrLen(szScoreInteger));
	FntSetFont(fntSav);

	// See if it fits with full text
	if (kcxStat - kcxBorder - cx - kcxBorder - cxScore - kcxBorder < 0) {

		// See if it fits without "Score" text
		if (kcxStat - kcxBorder - cx - kcxBorder - cxScoreInteger - kcxBorder < 0) {

			fnt = stdFont;
			fntSav = FntSetFont(fnt);
			cx = FntCharsWidth(sz, StrLen(sz));
			cxScore = FntCharsWidth(szScore, StrLen(szScore));
			cxScoreInteger = FntCharsWidth(szScoreInteger, StrLen(szScoreInteger));
			FntSetFont(fntSav);

			// See if it fits with full text
			if (kcxStat - kcxBorder - cx - kcxBorder - cxScore - kcxBorder < 0) {
				int cxMoves, cxEllipsis, cxAllowed;
				Int16 cch, cchT;
				Boolean fIgnore;

				// Doesn't fit with full text, lop off "Score"
				StrCopy(szScore, szScoreInteger);
				cxScore = cxScoreInteger;

				// Draw as much of the text that'll fit. Put ... in the title,
				// always show the # moves

				cxMoves = FntCharsWidth(szMoves, StrLen(szMoves));
				cxEllipsis = FntCharsWidth(gszEllipsis, StrLen(gszEllipsis));
				cxAllowed = kcxStat - kcxBorder - cxMoves - cxEllipsis - kcxBorder - cxScore - kcxBorder;
				fIgnore = true;
				cch = StrLen(szTitle);
				cchT = cch;
				FntCharsInWidth(szTitle, (Int16 *)&cxAllowed, (Int16 *)&cch, &fIgnore);

				// cch has been updated with the # of chars that will fit. Terminate
				// here and add ellipsis if there is enough room

				if (cch != cchT) {
					szTitle[cch] = 0;
					if (sizeof(szTitle) - cch >= StrLen(gszEllipsis))
						StrCopy(&szTitle[cch], gszEllipsis);
				}

				StrCopy(sz, szTitle);
				StrCat(sz, szMoves);
				fntSav = FntSetFont(fnt);
				cx = FntCharsWidth(sz, StrLen(sz));
				FntSetFont(fntSav);
			}
		} else {
			// It does fit without "Score" text

			StrCopy(szScore, szScoreInteger);
			cxScore = cxScoreInteger;
		}
	}

	// Fill background appropriately
	// If >= OS 3.5, use UI colors, otherwise black and white

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &dwVer);
	if (dwVer < 0x03503000) {
		rgbBack = 0x00000000;
		rgbFore = 0x00ffffff;
	} else {
		UIColorGetTableEntryRGB(UIFormFrame, (RGBColorType *)&rgbBack);
		((RGBColorType *)&rgbBack)->index = WinRGBToIndex((RGBColorType *)&rgbBack);
		UIColorGetTableEntryRGB(UIFormFill, (RGBColorType *)&rgbFore);
		((RGBColorType *)&rgbFore)->index = WinRGBToIndex((RGBColorType *)&rgbFore);
	}

	// Erase and draw

	WinSetColors((RGBColorType *)&rgbFore, (RGBColorType *)&rgbForeSav, (RGBColorType *)&rgbBack, (RGBColorType *)&rgbBackSav);
	fntSav = FntSetFont(fnt);
	cy = FntCharHeight() + 2;
	rc.topLeft.x = 0;
	rc.topLeft.y = 0;
	rc.extent.y = cy;
	rc.extent.x = 160;
   WinEraseRectangle(&rc, 0);

	if (dwVer < 0x03503000) {
		WinDrawInvertedChars(sz, StrLen(sz), kcxBorder, 1);
		WinDrawInvertedChars(szScore, StrLen(szScore), kcxStat - kcxBorder - cxScore, 1);
	} else {
		WinDrawChars(sz, StrLen(sz), kcxBorder, 1);
		WinDrawChars(szScore, StrLen(szScore), kcxStat - kcxBorder - cxScore, 1);
	}

	// Draw rounded corners

	rc.extent.y = 1;
	rc.extent.x = 2;
   WinDrawRectangle(&rc, 0);
	rc.extent.y = 2;
	rc.extent.x = 1;
   WinDrawRectangle(&rc, 0);   
	rc.topLeft.x = 158;
	rc.extent.y = 1;
	rc.extent.x = 2;
   WinDrawRectangle(&rc, 0);
	rc.topLeft.x = 159;
	rc.extent.y = 2;
	rc.extent.x = 1;	
   WinDrawRectangle(&rc, 0);

	// Restore

	FntSetFont(fntSav);
	WinSetColors((RGBColorType *)&rgbForeSav, NULL, (RGBColorType *)&rgbBackSav, NULL);
}

//-----------------------------------------------------------------------------
void WriteAboutInfo() {
   Int8              text [40];
   UInt8             idx, x;
   RectangleType     r;


   // copy 5 letters of Vexed to top of screen
   idx = BitmapIntroOffset + 8;           

   // copy all of bitmap
   r.topLeft.x = r.topLeft.y = 0;
   r.extent.x = BlockSizeX;
   r.extent.y = BlockSizeY;

   for (x = 17; x < 142; x += 25) {
      // Copy the source window (contains the image to draw) to the draw window.
      WinCopyRectangle (ObjectWindowHandles[idx++], 0, &r, x, 12, winPaint);
   }

   StrPrintF(text, "by the");
   WinDrawChars(text, StrLen(text), 67, 58);

   FntSetFont(1);
   StrPrintF(text, "Vexed SourceForge Project");
   WinDrawChars(text, StrLen(text), 12, 73);
   FntSetFont(0);

   StrPrintF(text, "A GPL Open Source Project");
   WinDrawChars(text, StrLen(text), 27, 88);

   StrPrintF(text, "vexed.sourceforge.net");
   WinDrawChars(text, StrLen(text), 35, 103);

   StrPrintF(text, "Version 2.0 Beta 4");
   WinDrawChars(text, StrLen(text), 44, 118);
}

//-----------------------------------------------------------------------------
void DrawIntro() {
   Int8              x, y;
   RectangleType     r;

   //intro screen moves
   MoveType moves[10] = {
      {5,2, Left, 1},
      {4,2, Left, 1},
      {0,0, Right, 1},
      {5,2, Left, 2},
      {3,3, Left, 1},
      {2,0, Left, 1},
      {4,0, Right, 1},
      {5,2, Left, 2},
      {3,3, Left, 1},
      {2,1, Right, 1}
   };

   // erase the whole screen first (Air)
   r.topLeft.x = 0;
   r.topLeft.y = 0;
   r.extent.x = 160;
   r.extent.y = 160;
   WinEraseRectangle(&r, 0);

   for (y = 0; y < sizeYI; y++) {
      for (x = 0; x < sizeXI; x++) {
         if (currentI[y][x] != AirI) {
            DrawBMIntro(currentI[y][x], (x * BlockSizeX) + 5, y * BlockSizeY);
         }
      }
   }

   for (x = 0; x < 10; x++) {
      PlayerMakeMoveI(moves[x].x, moves[x].y, moves[x].direction, moves[x].distance);
      if (EvtSysEventAvail(false)) {
         return;
      }
   }
   // erase the walls
   EraseWallsI();

   // move to top center
   r.topLeft.x = 5;
   r.topLeft.y = 120;
   r.extent.x = 125;
   r.extent.y = 40;
   WinEraseRectangle(&r, 0);

   WriteAboutInfo();
   return;
}

//-----------------------------------------------------------------------------
void EndSequence() {
   Int8              text [15];
   Int16             x1, x2, randnum1, randnum2;
   Int16             penx, peny;
   Boolean           bstate;
   RectangleType     r;

   while (bstate) EvtGetPen(&penx, &peny, &bstate);

   // draw some text of adulation to the rather smart user!
   r.topLeft.x = 0;
   r.topLeft.y = 16;
   r.extent.x = 160;
   r.extent.y = 128;
   WinEraseRectangle(&r, 0);

   FntSetFont(2);
   StrPrintF(text, "Well Done!");
   WinDrawChars(text, StrLen(text), 50, 75);
   FntSetFont(0);
   StrPrintF(text, "Game Complete");
   WinDrawChars(text, StrLen(text), 46, 90);

   r.extent.x = BlockSize;
   r.extent.y = BlockSize;
   // fire some random blocks along the top and bottom of the screen!
   while (1) {
      randnum1 = (Int16) ((SysRandom(0) / 4096) + 1);
      randnum2 = (Int16) ((SysRandom(0) / 4096) + 1);
      SoundEffect(HighBeep);
      for (x1 = 0; x1 <= 160; x1 ++) {
         r.topLeft.x = x1 - 1;
         r.topLeft.y = 16;
         WinEraseRectangle(&r, 0);
         DrawBM(randnum1 - 1, x1, 16, false);
         //WinDrawBitmap(Bitmap[randnum1], x1, 16);

         x2 = (160 - x1);

         r.topLeft.x = x2 + 1;
         r.topLeft.y = 128;
         WinEraseRectangle(&r, 0);
         DrawBM(randnum2 - 1, x2, 128, false);
         //WinDrawBitmap(Bitmap[randnum2], x2, 128);

         SysTaskDelay(2);
         EvtGetPen(&penx, &peny, &bstate);
         if (bstate)
            break;
      }
      SoundEffect(LowBeep);
      r.topLeft.x = 0;
      r.topLeft.y = 128;
      WinEraseRectangle(&r, 0);
      if (bstate)
         break;
   }
   SoundEffect(HighSweepLow);
   return;
}

//-----------------------------------------------------------------------------
void BlockCheck(UInt8 x, UInt8 y) {
   Int8              thisBlock;              // what kind of block is the pen on?
   RectangleType     r;

   if (InBlockCheck) {                       // already running?
      FrmAlert(AlertDebug1);                 // should never get here  $$TEMP$$
      return;
   }

   // got to here, we are starting a valid block check

   thisBlock = current[y][x];

   // erase blocks not like this one
   for (y = 0; y < sizeY; y ++) {
      for (x = 0; x < sizeX; x ++) {
         if ((current[y][x] != Air) && (current[y][x] != Wall) && (current[y][x] != thisBlock)) {
            r.topLeft.x = (x * 16);
            r.topLeft.y = ((y + 1) * 16);
            r.extent.x = 16;
            r.extent.y = 16;
            WinEraseRectangle(&r, 0);        // erase the on screen block
         }
      }
   }

   // Signify block check mode is on
   InBlockCheck = true;

   return;   
}

//-----------------------------------------------------------------------------
void DrawBM2(Int16 bitmapID, UInt8 x, UInt8 y, int ddx, int ddy, Boolean isId) {
   RectangleType     r;
   Int8              id;

   if (isId) {
      // translate resource ID into index into the windowhandle array
      id = (Int8) (bitmapID - firstBitmap + 8);  
   } else {
      id = (Int8) bitmapID;
   }

   if (isId || bitmapID == 8) {
      // copy entire window, one square BlockSize
      r.topLeft.x = r.topLeft.y = 0;
      r.extent.x = r.extent.y = BlockSize;
   } else {
      // Blocks have a 1 pixel border that is used for erasing when a block
      // is animating.
      r.topLeft.x = r.topLeft.y = 1;
      r.extent.x = r.extent.y = BlockSize;
      if (ddx == 1) {
         r.topLeft.x -= 1;
         r.extent.x += 1;
         x--;
      }
      if (ddx == -1)
         r.extent.x += 1;
      if (ddy == 1) {
         r.topLeft.y -= 1;
         r.extent.y += 1;
         y--;
      }
      if (ddy == -1)
         r.extent.y += 1;
   }

   ErrFatalDisplayIf (ObjectWindowHandles[id] == 0, "Unhandled object image");

   // Copy the source window (contains the image to draw) to the draw window.
   WinCopyRectangle (ObjectWindowHandles[id], 0, &r, x, y, winPaint);
}

//-----------------------------------------------------------------------------
void DrawBM(Int16 bitmapID, UInt8 x, UInt8 y, Boolean isId) {
   DrawBM2(bitmapID, x, y, 0, 0, isId);
}

//-----------------------------------------------------------------------------
void DrawControlIcons() {
   if (!InSolution) {
      // draw graphic controls at bottom of screen
      if (colorOk && VexedPreferences.ColorIcons) {
         DrawBM(BitmapRestart, 144, 144, true);
         DrawBM(BitmapBack, 16, 144, true);
         DrawBM(BitmapForward, 48, 144, true);
         DrawBM(BitmapSelectLevel, 32, 144, true);
         DrawBM(BitmapFirst, 0, 144, true);
         DrawBM(BitmapLast, 64, 144, true);
         DrawBM(BitmapMemory, 128, 144, true);
      } else {
         DrawBM(BitmapRestartM, 144, 144, true);
         DrawBM(BitmapBackM, 16, 144, true);
         DrawBM(BitmapForwardM, 48, 144, true);
         DrawBM(BitmapSelectLevelM, 32, 144, true);
         DrawBM(BitmapFirstM, 0, 144, true);
         DrawBM(BitmapLastM, 64, 144, true);
         DrawBM(BitmapMemoryM, 128, 144, true);
      }
   }
   return;
}

//-----------------------------------------------------------------------------
void UpdateBlocks(void) {
   Int8              y,x;                    // what kind of block is the pen on?
   RectangleType     r;

   r.extent.x = BlockSize;
   r.extent.y = BlockSize;

   // changed block sets, so redraw all non-wall blocks
   for (y = 0; y < sizeY; y ++) {
      for (x = 0; x < sizeX; x ++) {
         if ((current[y][x] != Air) && (current[y][x] != Wall)) {
            DrawBM(current[y][x] - 1, x * BlockSize, y * BlockSize + 16, false);
         }
         if (current[y][x] == Air) {
            r.topLeft.x = x * BlockSize;
            r.topLeft.y = y * BlockSize + 16;
            WinEraseRectangle(&r, 0);
         }
      }
   }

   return;   
}


