/*  Vexed v2.0 Beta - Vexed.c "The Main Stuff, i.e. event handlers, loops, etc"
    Copyright (C) 1999 James McCombe (cybertube@earthling.net)
    
    August 17, 2001 - Version 2.0 changes by Mark Ingebretson
    (ingebret@yahoo.com) and others.
    
    This file is part of Vexed.

    Vexed is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Vexed is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Vexed; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "PalmOS.h"

#pragma pack(2)

#include "Vexed.h"
#include "protos.h"

#define repeatDelta			(SysTicksPerSecond() / 4)	/* 0.5 seconds */

static Int16 StartApplication(void);
static void EventLoop(void);
static void StopApplication(void);
static Boolean frmMainEventH(EventPtr event);
static Boolean frmAboutEventH(EventPtr event);
static Boolean frmPrefsEventH(EventPtr event);
static Boolean frmInfoEventH(EventPtr event);
static Boolean frmIntroEventH(EventPtr event);
static Boolean frmSolutionEventH(EventPtr event);
void SetupButton(UInt16 controlId, Boolean upButton, Boolean turnedOn);
static Boolean LoadLastLevel();

struct {
   struct {
      Int16 x, y;
   } begin, end;
} playermove;

extern Preferences   VexedPreferences;
extern PlayGround    current;
extern PlayGround    stored;
extern Int8          CheatCount;
extern PlayGround    undo[UndoSize];
extern UInt8         UndoCount;
extern UInt8         UndoIndex;
extern Boolean       memorized;
extern UInt8         ReplayCurrent, 
                     ReplayPrevious;
extern MoveType      replay[2][ReplaySize];  // First row is previous moves, second row is current level
extern UInt8         ReplayCurrent,          // counts of moves  
                     ReplayPrevious; 
extern UInt16        cardNo;                 // for OS calls
extern LocalID       appID;                  // for OS calls
extern Boolean       colorOk;                // 8 bit color supported?
extern Boolean       refreshIcons;           // after preferences panel saved, refresh control icons
extern Int16         selectLevel;            // temp var for select level form
extern Int16         selectLevelRepeatCnt;   // Count of number of repeat events on select level form
extern Int8          selectDelta;            // count to add each button press
extern Boolean       cheatOn;
extern Boolean       InBlockCheck;           // are we in a block check
extern MemHandle	   ObjectBitmapHandles[BitmapCount];
extern BitmapPtr	   ObjectBitmapPtr[BitmapCount];
extern WinHandle	   ObjectWindowHandles[BitmapCount];
extern Boolean       InSolution;             // Are we displaying a solution
extern Boolean       InSolutionBack;         // are we doing a back move in solution?
extern Int16         MoveNum;                // Which move of solution are we now displaying
extern Boolean       SolutionRun;            // Is solution in free-running mode
extern Int8          OldSet;                 // Save off previous block set num
extern char          *MonoListP[];           // monochrome popup list has to be global

char gszEllipsis[4];
char stringSolution[512];
Int16 MoveCountMadeStored;
Int16 MoveCountParStored;

//---------------------------------------------------------------------------
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags) {
   Int16             error;
   if (cmd == sysAppLaunchCmdNormalLaunch) {
      error = StartApplication();            // Application start code
      if (error)
         return error;
      EventLoop();                           // Event loop
      StopApplication ();                    // Application stop code
   }

   return 0;
}

//-----------------------------------------------------------------------------
void LoadBitmaps(Boolean LoadAll) {
   WinHandle         oldDrawWinH;            // save off old draw handle
   UInt8             i;
   Err               err;
   UInt8             idx;
   RectangleType     r;
   Int8              bmCount;

   if (colorOk) {
      bmCount = BitmapCount;
   }
   else {
      bmCount = BitmapColorMono;                // load only bitmaps which are bitmap families
                                                // with both color and mono
                                                // exclude color only bitmaps
   }

   // Keep the Object graphics locked because they are frequently used
   oldDrawWinH = WinGetDrawWindow();
   r.topLeft.x = r.topLeft.y = 0;
   if (LoadAll) {
      for (i = 0; i < bmCount; i++) {
         int x, y;
         int cx, cy;

         // 8 block bitmaps are at the front of the array, followed by the wall
         idx = i + 8;
         ObjectBitmapHandles[idx] = DmGetResource( bitmapRsc, firstBitmap + i);
         ObjectBitmapPtr[idx] = MemHandleLock(ObjectBitmapHandles[idx]);
   
         // Make erasing border around intro bitmaps
         x = 0;
         y = 0;
         cx = 0;
         cy = 0;
         if (firstBitmap + i >= BitmapIntroV && firstBitmap + i <= BitmapIntroW) {
            x = 1;
            y = 1;
            cx = 2;
            cy = 2;
         }

         ObjectWindowHandles[idx] = WinCreateOffscreenWindow(ObjectBitmapPtr[idx]->width + cx, 
                                                           ObjectBitmapPtr[idx]->height + cx,
                                                           screenFormat, &err);
         WinSetDrawWindow(ObjectWindowHandles[idx]);
         r.extent.x = ObjectBitmapPtr[idx]->width + cx;
         r.extent.y = ObjectBitmapPtr[idx]->height + cx;
         WinEraseRectangle(&r, 0);
         
         WinDrawBitmap(ObjectBitmapPtr[idx], x, y);     
      }
   }

   for (i = 0; i < 8; i++) {
      ObjectBitmapHandles[i] = DmGetResource( bitmapRsc, BitmapBS_Extras + (VexedPreferences.BlockSet * 10) + i);
      ObjectBitmapPtr[i] = MemHandleLock(ObjectBitmapHandles[i]);
   
      ObjectWindowHandles[i] = WinCreateOffscreenWindow(ObjectBitmapPtr[i]->width + 2, 
                                                        ObjectBitmapPtr[i]->height + 2,
                                                        screenFormat, &err);
      WinSetDrawWindow(ObjectWindowHandles[i]);
      r.extent.x = ObjectBitmapPtr[i]->width + 2;
      r.extent.y = ObjectBitmapPtr[i]->height + 2;
      WinEraseRectangle(&r, 0);
         
      WinDrawBitmap(ObjectBitmapPtr[i], 1, 1);     
   }
   
   // restore old draw window
   WinSetDrawWindow(oldDrawWinH);

}

//-----------------------------------------------------------------------------
void FreeBitmaps(Boolean fFreeAll)
{
   UInt8             i;
   Int8              bmCount;
   UInt8             count;

   bmCount = BitmapColorMono;
   if (colorOk) {
      bmCount = BitmapCount;
   }

   if (fFreeAll) {
      count = bmCount + 8;
   } else {
      count = 8;
   }

   // Unlock and release the locked bitmaps
   for (i = 0; i < count; i++) {
      MemPtrUnlock(ObjectBitmapPtr[i]);
      DmReleaseResource(ObjectBitmapHandles[i]);

      if (ObjectWindowHandles[i])
         WinDeleteWindow(ObjectWindowHandles[i], false);
   }
}

//-----------------------------------------------------------------------------
static Int16 StartApplication(void) {
   UInt32            depth;
   UInt32            romVersion;
//   Int16             CurrentLevel;
//   Int16             SolvedLevels;
   
   colorOk = false;                          // default to no color
   
   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
   if (romVersion >=  sysMakeROMVersion(3, 5, 0, sysROMStageRelease, 0)) {
      // determine whether 8 bit color is supported
      WinScreenMode(winScreenModeGetSupportedDepths, NULL, NULL, &depth, NULL);
      if (depth & 0x80) {
         // color supported
         colorOk = true;
      }
   }

   // Get Ellipsis chars figured out. In OS 3.3 and above
   // there is a special ellipsis character.

   if (romVersion >= 0x03303000) {
      gszEllipsis[0] = '\30';
      gszEllipsis[1] = 0;
   } else {
      gszEllipsis[0] = '.';
      gszEllipsis[1] = '.';
      gszEllipsis[2] = '.';
      gszEllipsis[3] = 0;
   }

   // load preferences
   LoadPreferences();

   // WinDrawBitmap maps foreground / background color to UIColors when loading bitmaps.
   // This affects color bitmaps differently than B&W bitmaps, of which Vexed
   // has a mix. In addition, the color bitmaps have antialiasing built into them which
   // is assuming mixing into a white background. To circumvent these problems, we're
   // setting the background colors to white.

	if (romVersion >= 0x03503000) {
      RGBColorType rgbc;
      rgbc.index = 0;
      rgbc.r = 0xff;
      rgbc.g = 0xff;
      rgbc.b = 0xff;
      rgbc.index = WinRGBToIndex(&rgbc);
		UIColorSetTableEntry(UIDialogFill, &rgbc);
		UIColorSetTableEntry(UIFormFill, &rgbc);
      WinSetBackColor(rgbc.index);
	}

   // load all of the bitmaps
   LoadBitmaps(true);

   // attempt to load a last played level
   if (!LoadLastLevel()) {
      FrmGotoForm(FormLevelPack);
   }
   else if (VexedPreferences.SkipIntro) {
      FrmGotoForm(FormMain);
   }
   else {
      FrmGotoForm(FormIntro);
   }
}

//-----------------------------------------------------------------------------
Boolean LoadLastLevel()
{
   int x, y;
   int n;
   Boolean fLoaded;
   LevelPackPrefs *pLevelPack = &VexedPreferences.LevelPack[0];

   // If the pref board is valid, we have nothing to do.

   if (pLevelPack->LevelValid) {
      CLevelToLevel(pLevelPack->Level, current);
      return true;
   }

   // attempt to load one of the previous levels
   fLoaded = false;
   for (n = 0; n < MaxLevelPackPrefs; n++) {
      if (LoadLevel(VexedPreferences.LevelPack[n].DbName, VexedPreferences.LevelPack[n].CurrentLevel, true)) {
         fLoaded = true;
         break;
      }
   }
   
   // if failed to load any level, try the classic levels
   if (!fLoaded) {
      // force default level pack
      if (LoadLevel(DefaultLevelPackName, 0, true))
         fLoaded = true;
   }

   // return if this succeeded.
   return fLoaded;
}

//-----------------------------------------------------------------------------
static void EventLoop(void) {
   short             err;
   Int16             formID;
   FormPtr           form;
   EventType         event;

   do {
      EvtGetEvent(&event, 200);
      if (SysHandleEvent(&event)) {
         continue;
      }
      if (MenuHandleEvent((void *)0, &event, &err)) {
         continue;
      }
      if (event.eType == frmLoadEvent) {
         formID = event.data.frmLoad.formID;
         form = FrmInitForm(formID);
         FrmSetActiveForm(form);
         switch (formID) {
         case FormMain:
            FrmSetEventHandler(form, (FormEventHandlerPtr) frmMainEventH);
            break;

         case FormAbout:
            FrmSetEventHandler(form, (FormEventHandlerPtr) frmAboutEventH);
            break;

         case FormPrefs:
            FrmSetEventHandler(form, (FormEventHandlerPtr) frmPrefsEventH);
            break;

         case FormInfo:
            FrmSetEventHandler(form, (FormEventHandlerPtr) frmInfoEventH);
            break;

         case FormIntro:
            FrmSetEventHandler(form, (FormEventHandlerPtr) frmIntroEventH);
            break;
         
         case FormSolution:
            FrmSetEventHandler(form, (FormEventHandlerPtr) frmSolutionEventH);
            break;
         }
      }
      FrmDispatchEvent(&event);
   } while (event.eType != appStopEvent);
   return;
}

//-----------------------------------------------------------------------------
static void StopApplication(void) {
   FreeBitmaps(true);
   SavePreferences();
   FrmCloseAllForms();
}

//-----------------------------------------------------------------------------
void NextLevel(void) {
   Int16             i;

	// integrate move score
	if (VexedPreferences.LevelPack[0].CurrentLevel < kcParRecords) {
		VexedPreferences.LevelPack[0].Score[VexedPreferences.LevelPack[0].CurrentLevel].MoveCountMade = (UInt8)VexedPreferences.LevelPack[0].MoveCountMade;
		VexedPreferences.LevelPack[0].Score[VexedPreferences.LevelPack[0].CurrentLevel].MoveCountPar = (UInt8)VexedPreferences.LevelPack[0].MoveCountPar;
	} else {
		VexedPreferences.LevelPack[0].BaseMoveCountMade += VexedPreferences.LevelPack[0].MoveCountMade;
		VexedPreferences.LevelPack[0].BaseMoveCountPar += VexedPreferences.LevelPack[0].MoveCountPar;
	}
	VexedPreferences.LevelPack[0].MoveCountMade = 0;
	VexedPreferences.LevelPack[0].MoveCountPar = 0;
	
   // move onto next level if current is completed
   VexedPreferences.LevelPack[0].CurrentLevel ++;
   if (VexedPreferences.LevelPack[0].CurrentLevel > VexedPreferences.LevelPack[0].SolvedLevels + 1) {
      VexedPreferences.LevelPack[0].SolvedLevels = VexedPreferences.LevelPack[0].CurrentLevel - 1;
   }
   LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
   ReplayPrevious = ReplayCurrent;
   for (i=0; i < ReplayCurrent; i++) {
      replay[ReplayPIndex][i] = replay[ReplayCIndex][i];
   }
   ReplayCurrent = 0;

   SavePreferences();
   return;
}

//-----------------------------------------------------------------------------
static Boolean frmIntroEventH(EventPtr event) {
   FormPtr           form;
   Int16             handled = 0;
   Err               err;

   switch (event->eType) {
   case frmOpenEvent:
      // flush key/pen queues
      err = EvtFlushKeyQueue();
      err = EvtFlushPenQueue();
      form = FrmGetActiveForm();

      FrmDrawForm(form);
      FrmHideObject(form, FrmGetObjectIndex(form, ButtonOK3));
      DrawIntro();                           // draws the intro screen upon game startup
      if (EvtSysEventAvail(false)) {         // they want to quit early
         FrmGotoForm(FormMain);
      }
      else {
         FrmShowObject(form, FrmGetObjectIndex(form, ButtonOK3));
      }
      handled = 1;
      break;
   case ctlSelectEvent:
      if (event->data.ctlEnter.controlID == ButtonOK3) {
         FrmGotoForm(FormMain);
         handled = 1;
      }
   }
   return handled;
}
//-----------------------------------------------------------------------------
void ExitSolution() {
	CopyLevel(current,stored);
	VexedPreferences.LevelPack[0].MoveCountMade = MoveCountMadeStored;
	VexedPreferences.LevelPack[0].MoveCountPar = MoveCountParStored;
	InSolution = false;
	stored[0][0] = 0;								// clear memorized area
	memorized = false;
	FrmReturnToForm(0);
	UpdateStats();
}

//-----------------------------------------------------------------------------
void MakeSolutionMove(UInt8 BackNum) {
	int x, y;
	char chX, chY;
	int count;
	Int8 direction;

   // if in back function we know it won't be the end
   if (InSolutionBack) {
      // Get the move
   
   
      chX = stringSolution[BackNum * 2];
      chY = stringSolution[BackNum * 2 + 1];

   }
   else {   
	   // End of the solution?

	   count = StrLen(stringSolution);
	   if (MoveNum * 2 >= count) {
		   ExitSolution();
	   	return;
   	}
      // Get the move
   
   
      chX = stringSolution[MoveNum * 2];
      chY = stringSolution[MoveNum * 2 + 1];
   }
   
	x = chX - 'a';
	if (chX <= 'Z') {
		direction = Left;
		x = chX - 'A';
	}
	y = chY - 'a';
	if (chY <= 'Z') {
		direction = Right;
		y = chY - 'A';
	}

	// Check for validity

	if (x < 0 || x >= sizeX || y < 0 || y >= sizeY) {
		ExitSolution();
		return;
	}

	// Add the event
   if (InSolutionBack) {                     // if doing a solution back event, just do the move
      (void) PlayerMakeMove(x, y, direction, 1);
   }
   else {                                    // otherwise queue it
      AddMoveEvent(x, y, direction);
      MoveNum++;
   }
	return;
}

//-----------------------------------------------------------------------------
static Boolean frmSolutionEventH(EventPtr event) {
   FormPtr           form;
   Int16             handled = 0;
   Err               err;
   Boolean           tmp;
   RectangleType     r;
	Int8					tempS[40];
   Int16             penx, peny;
   Boolean           bstate;
   Int8              i;

   switch (event->eType) {
   case frmOpenEvent:
      InSolution = true;
      InSolutionBack = false;
      SolutionRun = true;                    // free-running until they tap
      MoveNum = 0;

      // memorize state of current level
      CopyLevel(stored,current);
		MoveCountMadeStored = VexedPreferences.LevelPack[0].MoveCountMade;
		MoveCountParStored = VexedPreferences.LevelPack[0].MoveCountPar;

      // flush key/pen queues
      err = EvtFlushKeyQueue();
      err = EvtFlushPenQueue();
      form = FrmGetActiveForm();

      FrmDrawForm(form);
      if (colorOk && VexedPreferences.ColorIcons) {
         DrawBM(BitmapRestart, 144, 144, true);
      }
      else {
         DrawBM(BitmapRestartM, 144, 144, true);
      }
      LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);

      // add the first move into the queue
		MakeSolutionMove(0);            
      handled = 1;
      break;
   case ctlSelectEvent:
      if (event->data.ctlEnter.controlID == ButtonDone) {
         ExitSolution();
         handled = 1;
      }
      break;
   case keyDownEvent:
      if (SolutionRun) {                     // we were running in automatic mode
         SolutionRun = false;			
      }
      else if (event->data.keyDown.chr == pageDownChr) {   // page up key pressed?
         MakeSolutionMove(0);
      }
      else if (event->data.keyDown.chr == pageUpChr) {   // page down key pressed?
         goto SolutionBack;
      }
      handled = 1;
      break;
   case MakeMoveEvent:
      if (PlayerMakeMove(event->data.generic.datum[0], event->data.generic.datum[1], 
                           event->data.generic.datum[2], 1)) {
			// all done

			//ExitSolution();

         // cycle back to the beginning instead of exiting
         SolutionRun = false;
         LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
         MoveNum = 0;
		}
		else if (SolutionRun) {

			// don't queue another if there's a pen event out there
			if (!EvtSysEventAvail(false)) {     // they want to quit early
				MakeSolutionMove(0);					// if all done				
			}
		}
      handled = 1;
      break;
	case penUpEvent:
      err = EvtFlushPenQueue();					// flush any other pen events
      EvtGetPen(&penx, &peny, &bstate);
      if (SolutionRun) {                     // we were running in automatic mode
         SolutionRun = false;			
         // did they press the done button?
         if ((penx >5) && (penx < 44) && (peny > 147) && (peny < 159)) {
            MakeSolutionMove(0);
         }
      }
      else {
			// did they tap on a control icon at the bottom?
			if (peny > 143) {                   // command icon area?
				// ignore
				if (penx < 70) {
				}
				// back
				else if (penx < 80) {
SolutionBack:
               if (MoveNum > 0) {
                  InSolutionBack = true;
                  LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
                  for (i = 0; i < MoveNum - 1; i++) {
   //ErrDisplay("making back move");
                     MakeSolutionMove(i);
                  }
                  UpdateBlocks();            // redraw the blocks
                  MoveNum--;
                  InSolutionBack = false;
               }
				}
				// step
				else if (penx < 90) {
					MakeSolutionMove(0);
				}
            // restart
            else if (penx > 143) {
               SolutionRun = true;
               LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
               MoveNum = 0;
					MakeSolutionMove(0);
            }
			}
      }
      handled = 1;
      break;
   }
   return handled;
}

//-----------------------------------------------------------------------------
void DoTitleSelect()
{
   EventType event;

   MemSet (&event, sizeof(EventType), 0);
   event.eType = keyDownEvent;
   event.data.keyDown.chr = menuChr;
   event.data.keyDown.keyCode = 0;
   event.data.keyDown.modifiers = commandKeyMask;
   EvtAddEventToQueue (&event);
}

//-----------------------------------------------------------------------------
static Boolean frmMainEventH(EventPtr event) {
   FormPtr           form;
   Int16             handled = 0;
   Int8              dist;   
   Int16             penx, peny;
   Boolean           bstate;
   UInt8             Button;
   RectangleType     r;
   UInt32            KeyStatus;
   Err               err;

   switch (event->eType) {
   
   case frmOpenEvent:
      form = FrmGetActiveForm();
      FrmDrawForm(form);

      // Draw current level
      DrawLevel();

      // Reset replay counts, other initialization
      ReplayCurrent = ReplayPrevious = 0;
      cheatOn = false;
      InBlockCheck = false;                  // are we in a block check?

      // draw graphic controls at bottom of screen
      DrawControlIcons();
      handled = 1;
      break;

   case frmUpdateEvent:                      // fix draw of screen after modal dialogs 
      form = FrmGetActiveForm();             
      FrmDrawForm(form);                     
      RedrawLevel();
      DrawControlIcons();
      if (UndoCount > 0) {
         if (colorOk && VexedPreferences.ColorIcons) {
            DrawBM(BitmapUndo,96,144,true);
         } else {
            DrawBM(BitmapUndoM,96,144,true);
         }
      }
      if (memorized) {                       // anything memorized? 
         if (colorOk && VexedPreferences.ColorIcons) {
            DrawBM(BitmapRecall, 112,144, true);  // load the recall icon 
         } else {
            DrawBM(BitmapRecallM, 112,144, true); // load the recall icon 
         }
      }
      handled = 1;                           
      break;                                 

   case penDownEvent:
      EvtGetPen(&penx, &peny, &bstate);
      if (peny < 16) {                       // title bar area? 
         DoTitleSelect();
         handled = 1;
         break;                              // return to OS 
      }
      if (peny < 144) {                      // the gameplay area
         playermove.begin.x = (Int16) (penx / BlockSize);
         playermove.begin.y = (Int16) ((peny / BlockSize) - 1);

         // first determine if in a valid block
         if ((current[playermove.begin.y][playermove.begin.x] != Air ) &&
             (current[playermove.begin.y][playermove.begin.x] != Wall)) {

            // ensure page up key is down
            if (KeyCurrentState() & keyBitPageUp) {
               BlockCheck(playermove.begin.x, playermove.begin.y);
            }
         }
         handled = 1;
         break;
      } else if (peny > 144) {               // the control icons area
         Button = penx >> 4;
         switch (Button) {

         //First level icon
         case FirstButton:
            if (VexedPreferences.LevelPack[0].CurrentLevel != 0) {
               VexedPreferences.LevelPack[0].CurrentLevel = 0;
               LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
            }
            CheatCount=0;
            handled = 1;
            break;

            // Previous level icon
         case PreviousButton:
            if (VexedPreferences.LevelPack[0].CurrentLevel != 0) {
               VexedPreferences.LevelPack[0].CurrentLevel --;
               LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
            }
            CheatCount=0;
            handled = 1;
            break;

            // Select level icon
         case SelectButton:
            DoSelectLevel();
            handled = 1;
            break;

            // Next level icon
         case NextButton:
            if ((VexedPreferences.LevelPack[0].SolvedLevels != -1) && (VexedPreferences.LevelPack[0].CurrentLevel <=  VexedPreferences.LevelPack[0].SolvedLevels) || ((cheatOn) && (CheatCount >= 4))) {
               VexedPreferences.LevelPack[0].CurrentLevel ++;
               LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
            } else {
               CheatCount++;
            }
            handled = 1;
            break;

            // Last level icon
         case LastButton:
            // if current != Solved
            if (VexedPreferences.LevelPack[0].SolvedLevels != -1) {
               if (VexedPreferences.LevelPack[0].CurrentLevel < VexedPreferences.LevelPack[0].SolvedLevels + 1) {
                  VexedPreferences.LevelPack[0].CurrentLevel = VexedPreferences.LevelPack[0].SolvedLevels + 1;
                  if (VexedPreferences.LevelPack[0].CurrentLevel == VexedPreferences.LevelPack[0].MaxLevels) {
                     VexedPreferences.LevelPack[0].CurrentLevel = VexedPreferences.LevelPack[0].MaxLevels - 1;
                  }
                  LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
               }
               CheatCount=0;
            }
            handled = 1;
            break;

            // Undo icon
         case UndoButton:
            if (UndoCount > 0) {
               UndoCount--;
               UndoIndex = (UndoIndex - 1 + UndoSize) % UndoSize;
               CopyLevel(current,undo[UndoIndex]);
               VexedPreferences.LevelPack[0].MoveCountMade += 1;
               RedrawLevel();
               if (UndoCount == 0) {
                  // clear the undo icon 
                  r.topLeft.x = 96;
                  r.topLeft.y = 144;
                  r.extent.x = BlockSize;
                  r.extent.y = BlockSize;
                  WinEraseRectangle(&r, 0);
               }
               // decrement the replay count
               ReplayCurrent--;
            }
            handled = 1;
            break;

            // Memorize icon
         case MemorizeButton:
            CopyLevel(stored,current);
				MoveCountMadeStored = VexedPreferences.LevelPack[0].MoveCountMade;
				MoveCountParStored = VexedPreferences.LevelPack[0].MoveCountPar;

            // display the recall icon
            if (colorOk && VexedPreferences.ColorIcons) {
               DrawBM(BitmapRecall, 112, 144, true);  // load the recall icon 
            } else {
               DrawBM(BitmapRecallM, 112, 144, true); // load the recall icon 

            }
            memorized = true;                

            handled = 1;
            break;

            // Recall icon
         case RecallButton:
            if (stored[0][0] != 0) {
               CopyLevel(current,stored);
					VexedPreferences.LevelPack[0].MoveCountMade = MoveCountMadeStored;
					VexedPreferences.LevelPack[0].MoveCountPar = MoveCountParStored;
               DrawLevel();
            }
            handled = 1;
            break;

            // Restart level icon
         case RestartButton:
            LoadLevel(NULL, VexedPreferences.LevelPack[0].CurrentLevel, false);
            ReplayCurrent = 0;
            handled = 1;
            break;
         }
         break;
      }

   case penUpEvent:
      // check to see if block check mode is now over
      if (InBlockCheck) {                   
         // end block check mode
         err = EvtFlushKeyQueue();           // flush key queue
         err = EvtFlushPenQueue();
         RedrawLevel();
         InBlockCheck = false;               // cleared for next time
         handled = 1;
         break;
      }
      EvtGetPen(&penx, &peny, &bstate);
      if (peny < 16) {                       // title bar area? 
         break;                              // return to OS 
      }
      if (peny < 144) {                      // penUpEvent is only an issue in the gameplay area
         playermove.end.x = (Int16) (penx / BlockSize);
         playermove.end.y = (Int16) ((peny / BlockSize) - 1);
         if (playermove.begin.y == playermove.end.y) {
            if (playermove.end.x > playermove.begin.x) {
               dist = (Int8) (playermove.end.x - playermove.begin.x);
               if (UndoCount < UndoSize) {
                  UndoCount++;
               }
               CopyLevel(undo[UndoIndex],current);
               UndoIndex = (UndoIndex + 1) % UndoSize;
               if (PlayerMakeMove2(playermove.begin.x, playermove.begin.y, Right, dist, true)) {
                  NextLevel();
                  handled = 1;
                  break;
               }
            } else if (playermove.end.x < playermove.begin.x) {
               dist = (Int8) (playermove.begin.x - playermove.end.x);
               if (UndoCount < UndoSize) {
                  UndoCount++;
               }
               CopyLevel(undo[UndoIndex],current);
               UndoIndex = (UndoIndex + 1) % UndoSize;
               if (PlayerMakeMove2(playermove.begin.x, playermove.begin.y, Left, dist, true)) {
                  NextLevel();
                  handled = 1;
                  break;
               }
            }
         }
      }
      handled = 1;
      break;

   case menuEvent:
      switch (event->data.menu.itemID) {
      
      case ItemAbout:
         FrmPopupForm(FormAbout);
         handled = 1;
         break;

		case ItemBeamVexed:
			BeamDatabase(cardNo, appID, "Vexed.prc", "Vexed!");
			break;

      case ItemSolution:
			// First try to get the solution. Perhaps there isn't one
			if (!GetLevelProperty(VexedPreferences.LevelPack[0].CurrentLevel, "solution", stringSolution, sizeof(stringSolution))) {
				FrmAlert(AlertNoSolution);
				handled = 1;
				break;
			}
			// Really bring up solution?
			if (FrmAlert(AlertShowSolution) != 0) {
				handled = 1;
				break;
			}
			// Add 5 to BaseMoveCountMade so that redoing the level doesn't
			// "forgive" these 5 points
			VexedPreferences.LevelPack[0].BaseMoveCountMade += 5;
         FrmPopupForm(FormSolution);
         handled = 1;
         break;

      case ItemPreferences:
         FrmPopupForm(FormPrefs);
         handled = 1;
         break;

      case ItemHowToPlay:
         FrmHelp(StringInfo);
         handled = 1;
         break;

         // clear levels back to 0
      case ItemClear:
         if (!FrmAlert(AlertResetYesNo)) {
            VexedPreferences.LevelPack[0].SolvedLevels = -1;     
            VexedPreferences.LevelPack[0].CurrentLevel = 0;
				VexedPreferences.LevelPack[0].BaseMoveCountMade = 0;
            LoadLevel(NULL, 0, false);
         }
         handled = 1;                              
         break;                                    

#if 0
         // skip to second set of levels, #59
      case ItemSkip:
         if (VexedPreferences.LevelPack[0].CurrentLevel < 59) {
            if (!FrmAlert(AlertNewLevelsYesNo)) {
               VexedPreferences.LevelPack[0].SolvedLevels = 58; 
               VexedPreferences.LevelPack[0].CurrentLevel = 59; 
               LoadLevel(59, false);
            }
         } else {
            FrmAlert(AlertSecondLevel);            
         }
         handled = 1;                              
         break;                                    
#endif

#if 0
         // add 59 for those people upgrading from version 1.4a 
         // (those levels were 0-58, now they're 59-117
      case ItemAdjust:
         if (VexedPreferences.LevelPack[0].SolvedLevels < 59) {
            if (!FrmAlert(AlertAdjustYesNo)) {
               VexedPreferences.LevelPack[0].SolvedLevels = VexedPreferences.LevelPack[0].SolvedLevels + 59;  
               VexedPreferences.LevelPack[0].CurrentLevel = VexedPreferences.CurrentLevel + 59;  
               LoadLevel(59, false);
            }
         } else {
            FrmAlert(AlertSecondLevel);      
         }
         handled = 1;                        
         break;                              
#endif

      case ItemLevelPacks:
         DoLevelPackForm();
         handled = 1;
         break;

         // Level Info 
      case ItemInfo:
         FrmPopupForm(FormInfo);
         handled = 1;
         break;

         // Replay current level
      case ItemReplayCurrent:
         if (!ReplayCurrent) {               // Current level replay data empty?
            FrmAlert(AlertNoReplayCurrent);      
         } else {
            ReplayLevel(ReplayCIndex);       // Replay current level
         }
         handled = 1;
         break;

         // Replay previous level
      case ItemReplayPrevious:
         if (VexedPreferences.LevelPack[0].CurrentLevel == 0) { // on the first level
            FrmAlert(AlertNoPreviousLevel);      
         } else if (!ReplayPrevious) {       // Previous level replay data empty?
            FrmAlert(AlertNoReplayPrevious);      
         } else {
            ReplayLevel(ReplayPIndex);       // Replay previous level
         }
         handled = 1;
         break;
      }

      // check for Graffiti character, page up key
   case keyDownEvent:
      if (event->data.keyDown.chr == 'c') {
         cheatOn = true;
         CheatCount = 0;
         handled = 1;
      }
      else if (event->data.keyDown.chr == pageUpChr) {   // page up key pressed?
         // find out if the pen is down and in a valid block
         // only do this on initial key press, not autorepeat ones
         if (!InBlockCheck && !(event->data.keyDown.modifiers & autoRepeatKeyMask)) {
            EvtGetPen(&penx, &peny, &bstate);
            if (bstate) {                          // if pen is down
               if ((peny > 15) && (peny < 144)) {     // in the game board?
                  playermove.begin.x = (Int16) (penx / BlockSize);
                  playermove.begin.y = (Int16) ((peny / BlockSize) - 1);
               
                  // first determine if in a valid block
                  if ((current[playermove.begin.y][playermove.begin.x] != Air ) &&
                     (current[playermove.begin.y][playermove.begin.x] != Wall)) {      
                     BlockCheck(playermove.begin.x, playermove.begin.y);
                  }
               }
            }
         }
         //err = EvtFlushKeyQueue();           // flush key queue
         handled = 1;
      }
      break;

   case nilEvent:
      handled = 1;
      break;
   }
   if (refreshIcons) {
      DrawControlIcons();
      if (colorOk && VexedPreferences.ColorIcons) {
         if (memorized) {                    // anything memorized? 
            DrawBM(BitmapRecall, 112, 144, true);// load the recall icon 
         }
         if (UndoCount > 0) {
            DrawBM(BitmapUndo, 96, 144, true);   
         }
      } else {
         if (memorized) {                    // anything memorized? 
            DrawBM(BitmapRecallM, 112, 144, true);// load the recall icon 
         }
         if (UndoCount > 0) {
            DrawBM(BitmapUndoM, 96, 144, true);   
         }
      }
      UpdateBlocks();                        // update blocks, too
      
      refreshIcons = false;
   }
   return handled;
}

//-----------------------------------------------------------------------------
static Boolean frmAboutEventH(EventPtr event) {
   FormPtr           form;
   Int16             handled = 0;

   switch (event->eType) {
   case frmOpenEvent:
      form = FrmGetActiveForm();
      FrmDrawForm(form);
      WriteAboutInfo();
      handled = 1;
      break;
   case ctlSelectEvent:
      if (event->data.ctlEnter.controlID == ButtonOK) {
         FrmReturnToForm(0);
         handled = 1;
         break;
      }
      else if (event->data.ctlEnter.controlID == ButtonCredits) {
         FrmHelp(StringCredits);
         handled = 1;
         break;
      }
   }
   return handled;
}

//-----------------------------------------------------------------------------
static Boolean frmInfoEventH(EventPtr event) {
   FormPtr           form;
   Int16             handled = 0;
   Int16             BlockCount[16];
   Int8              BlockCountStr[10][3];
   Int8              x,y,i;

   switch (event->eType) {
   case frmOpenEvent:
      form = FrmGetActiveForm();
      for (i=0;i < 16; i++) {
         BlockCount[i] = 0;
      }
      for (y = 0; y < sizeY; y ++) {
         for (x = 0; x < sizeX; x ++) {
            BlockCount[current[y][x]]++;     // increment count
         }
      }
      for (i=0;i < 10; i++) {
         StrPrintF(BlockCountStr[i], "%i", BlockCount[i]);
      }    

      FrmCopyLabel (form, LabelBlock1Count,BlockCountStr[1]);
      FrmCopyLabel (form, LabelBlock2Count,BlockCountStr[2]);
      FrmCopyLabel (form, LabelBlock3Count,BlockCountStr[3]);
      FrmCopyLabel (form, LabelBlock4Count,BlockCountStr[4]);
      FrmCopyLabel (form, LabelBlock5Count,BlockCountStr[5]);
      FrmCopyLabel (form, LabelBlock6Count,BlockCountStr[6]);
      FrmCopyLabel (form, LabelBlock7Count,BlockCountStr[7]);
      FrmCopyLabel (form, LabelBlock8Count,BlockCountStr[8]);

      FrmDrawForm(form);
      
      // now need to draw the blocks from memory to support multiple block sets
      DrawBM(0, 10, 20, false);
      DrawBM(1, 10, 36, false);
      DrawBM(2, 10, 52, false);
      DrawBM(3, 10, 68, false);
      DrawBM(4, 72, 20, false);
      DrawBM(5, 72, 36, false);
      DrawBM(6, 72, 52, false);
      DrawBM(7, 72, 68, false);

      handled = 1;
      break;
   case ctlSelectEvent:
      if (event->data.ctlEnter.controlID == InfoButtonOK) {
         FrmReturnToForm(0);
         handled = 1;
         break;
      }
   }
   return handled;
}

//-----------------------------------------------------------------------------
static Boolean frmPrefsEventH(EventPtr event) {
   FormPtr           form;
   Int16             handled = 0;
   ListType          *list;
   ControlType       *trigger;
   Char              *label;

   form = FrmGetActiveForm();
   list = FrmGetObjectPtr(form, FrmGetObjectIndex(form, PopupBlockSetList));
   
   switch (event->eType) {
   case frmOpenEvent:

      // set the check boxes and list to the current preferences state
      trigger = FrmGetObjectPtr(form, FrmGetObjectIndex(form, PopupBlockSet));
      label = (Char *) CtlGetLabel(trigger);
      *label = VexedPreferences.BlockSet + '0';
      CtlSetLabel(trigger, label);

      FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxPieceAnim), VexedPreferences.PieceMoveAnim);
      FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxGravityAnim), VexedPreferences.GravityAnim);
      FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxEliminationAnim), VexedPreferences.EliminationAnim);
      FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxBlindsEffect), VexedPreferences.BlindsEffect);
      FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxSound), VexedPreferences.Sound);
      FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxSkipIntro), VexedPreferences.SkipIntro);
      if (colorOk) {
         FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxColorIcons), VexedPreferences.ColorIcons);
      }

      // turn off the color icons checkbox in preferences
      if (!colorOk) {
         // scale back list for monochrome devices
         LstSetListChoices(list, MonoListP, 2);
         LstSetHeight(list, 2);
         LstMakeItemVisible(list, 1);
         FrmHideObject(form, FrmGetObjectIndex(form, CheckBoxColorIcons));
      }
      else {
         LstSetHeight(list, 5);
      }
      
      // set up the block set list
      LstSetSelection(list, VexedPreferences.BlockSet - 1);
      OldSet = VexedPreferences.BlockSet;

      // fall through
   case frmUpdateEvent:
      FrmDrawForm(form);
      // Draw a separator line
      WinDrawLine(6,70,150,70);

      handled = 1;
      break;
   case lstSelectEvent:
      switch (event->data.lstSelect.listID) {
      case PopupBlockSetList:
         // VexedPreferences.BlockSet = LstGetSelection(list) + 1;
         handled = 1;
      }
      break;
   case ctlSelectEvent:
      switch (event->data.ctlEnter.controlID) {
      case ButtonOK2:

         // set the preferences state to the state of the checkboxes on the form
         VexedPreferences.PieceMoveAnim = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxPieceAnim));
         VexedPreferences.GravityAnim = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxGravityAnim));
         VexedPreferences.EliminationAnim = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxEliminationAnim));
         VexedPreferences.BlindsEffect = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxBlindsEffect));
         VexedPreferences.Sound = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxSound));
         VexedPreferences.SkipIntro = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxSkipIntro));
         VexedPreferences.BlockSet = LstGetSelection(list) + 1;
         if (colorOk) {
            VexedPreferences.ColorIcons = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxColorIcons));
         }

         // if there is a new block set, load the bitmaps
         if (OldSet != VexedPreferences.BlockSet) {
            FreeBitmaps(false);
            LoadBitmaps(false);              // load the new bitmap set
         }

         // leave the preferences form
         refreshIcons = true;
         FrmReturnToForm(0);
         handled = 1;
         break;
      case ButtonCancel:
         FrmReturnToForm(0);
         handled = 1;
         break;
      }
      break;
   }
   return handled;
}

